-- ALU_16.vhd
-- version 20090102
-- version 20090108 : adding MUL8x4
-- version 20090526 : added RegSet, this is now more a datapath than an ALU
-- version 20090603 : changed std_logic to std_ulogic everywhere (it helped catch a few bugs)

-- This file gathers the separate units
-- and connects them so that latency is minimal.

--           _EN         _ENr                      _ENrr
-- ASU : MUL8x4---FF------------------------FFx---------------L8----\MUX0___Result
-- SR :        ---FF--[......extern......]--FF1-\_MUX2__m2__MUX1_m1_/(L9)
-- ROP2 :      ---FF------------L5-\_MUX4___FF2_/ (L7)      (L8)
-- SHL  :      ---FF------------L5-/ (L6)                   /
-- IE          ---FF---\MUX5-m5-IE---L2-----FF3-\_MUX3_m3__/
-- Mem          IOFF---/------------------------/

-- I should add : NPC, PC, 

library ieee;
use ieee.std_logic_1164.all;
library work;
use work.yasep_definitions;

entity datapath_16 is
--  generic (
--    see work.yasep_definitions
--  );
  port(
    Instruction : in SULV32; -- bit0 is the short/long flag
    MemRead, SR_Read, NPC : in SULVY;
    clk, new_instruction,
    SR_EN, MEM_ENr, MEM_ENrr, OffsetA,
    BypassRegA, BypassRegB, BypassRegA2, BypassRegB2 : in std_ulogic;
    AdrWr, BankID : in SULV4;

    MemWrite : out SULVY;
    Carry_out, ROP2_Combine : out std_ulogic);
end datapath_16;

architecture together of datapath_16 is
  signal ROP2_OUT, ResSHL, ResIE, FF1, FF2, FF3, DataAr, DataBr,
     m1, m2, m3, m5, DataA, DataB, DataAreg, DataBreg, Result,
      Imm4, Imm16 : SULVY;
  signal Sum : std_ulogic_vector(YASEP_SIZE downto 0);
  signal MulOutLo, MulOutHi : std_ulogic_vector(11 downto 0);
  signal MulMuxInput : SULV8;
  signal AdrRd1, AdrRd2 : SULV4;
  signal GroupBits, FunctionBits : SULV3;
  signal Form : SULV2;
  signal IE_enr,  SR_ENr,  add_ENr,  MulEnr, SHL_enr, ROP2_ENr,
         IE_enrr, SR_ENrr, add_ENrr, Loadr, SignExtr, OffsetAr,
      add_EN, MulEn, SHL_en, IE_en, ROP2_EN, MulWrEn,
      AddMulRop2, MulHiLo, Rotation, Direction, Arithmetic,
      Load, SignExt, ROP2_S1, ROP2_S2, ROP2_X2, Addsub, AuxImm4,
      R7WrEn, R7RdEn1, R7RdEn2 : std_ulogic;
  signal Opcode : std_ulogic_vector(R_OPCODE);
begin

    assert (YASEP_SIZE = 16)
      report "YASEP_SIZE must be 16 for this version."
      severity failure;

------------------------------------------------------------------------------------
-- First pipeline stage :
--  * Instruction(15 downto 0) is available,
--  * AdrRd1 and AdrRd2 are fed to the R7
--  * The read enable are computed (R7RdEn1, R7RdEn2)
--  * some high-fanout signals are created
------------------------------------------------------------------------------------

  process(clk, Instruction, new_instruction)
  begin
    if (clk'event and clk='1') then
      MemRead_r <= MemRead;
      if (new_instruction='1') then
        Instruction_r <= Instruction;
        NPC_r <= NPC;
      end if;
    end if;
  end process;

-- decoding : about 2 logic layers

-- Instruction fields :
  Opcode       <= Instruction(R_OPCODE);
  Form         <= Instruction(1 downto 0);
  GroupBits    <= Instruction(R_GROUP);
  FunctionBits <= Instruction(R_FUNCTION);
  AdrRd1       <= Instruction(R_SRC1);
  AdrRd2       <= Instruction(R_IMM4);
  Imm4(AdrRd2'range) <= AdrRd2;
  Imm4(Imm4'left downto 4) <= (others => AdrRd2(3)); -- sign extension
  Imm16(14 downto 0) <= Instruction(30 downto 16);
  Imm16(YASEP_SIZE-1 downto 15) <= (others => Instruction(31)); -- sign extension

-- Extended form :
  AuxImm4 <= Instruction(16);

-- Groups :
  IE_en   <= '1' when (GroupBits = GROUP_IE)   else '0';
  SHL_en  <= '1' when (GroupBits = GROUP_SHL)  else '0';
  ROP2_EN <= '1' when (GroupBits = GROUP_ROP2) else '0';
  MulEn   <= '1' when (GroupBits = GROUP_MUL) and not (YASEP_MUL = 0)  else '0';
  add_EN  <= '1' when (GroupBits = GROUP_ASU or  MulEn = '1') else '0';
  AddMulRop2 <= add_EN or ROP2_EN or MulEnr;

-- multiply table initialisation ?
  if (YASEP_MUL = 0) generate
    MulWrEn <= '0';
  end generate;
  if not(YASEP_MUL = 0) generate
    MulWrEn <= '1' when (Opcode = (F_MULI & GROUP_MUL)) else '0';
  end generate;

-- trick : reduce toggles and leave some unchanged values
  process(clk, GroupBits, FunctionBits)
    variable TmpDecode : SULV4;
  begin
    if (clk'event and clk='1') then
      case GroupBits is
        when GROUP_ASU =>
		  if (FunctionBits = F_ADD) then
            Addsub <= '0'; -- only one add function
          else
            Addsub <= '1'; -- comparison, min/max & sub => sub
          end if;
        when GROUP_ROP2 =>
          case FunctionBits is   --       ROP2_S1  ROP2_S2  ROP2_X2  Addsub
            when F_XOR  => TmpDecode := "0000"; -- "000"    0        0        0       0
            when F_XORN => TmpDecode := "0001"; -- "001"    0        0        0       1
            when F_OR   => TmpDecode := "X100"; -- "010"    X        1        0       0
            when F_ORN  => TmpDecode := "X101"; -- "011"    X        1        0       1
            when F_NOR  => TmpDecode := "X110"; -- "100"    X        1        1       0
            when F_AND  => TmpDecode := "1000"; -- "101"    1        0        0       0
            when F_ANDN => TmpDecode := "1001"; -- "110"    1        0        0       1
            when F_NAND => TmpDecode := "1010"; -- "111"    1        0        1       0
          end case;
          Addsub  <= TmpDecode(0);
          ROP2_S1 <= TmpDecode(1);
          ROP2_S2 <= TmpDecode(2);
          ROP2_X2 <= TmpDecode(3);
        when GROUP_MUL =>
          if (FunctionBits = F_MUL8H) and (not (YASEP_MUL = 0)) then
            MulHiLo <= '1';
          else
            MulHiLo <= '0';
          end if;
        when GROUP_SHL =>
          Direction  <= FunctionBits(0);
          Rotation   <= FunctionBits(1);
          Arithmetic <= FunctionBits(2);
        when GROUP_IE =>
		  SignExt <= FunctionBits(0);
          Load    <= FunctionBits(1);
		when others => TmpDecode := "XXXX"; -- do nothing
      end case;
    end if;
  end process;



------------------------------------------------------------------------------------
-- Second pipeline stage :
--  * read the register set and the 2 main operands
--  * some other decoding signals are also computed.
------------------------------------------------------------------------------------

R7:
  entity work.RegSet16(A3P)
  port map(
    Clk => clk,
    WrEn => R7WrEn, -- Low=Write
    RdEn1 => R7RdEN1,
    RdEn2 => R7RdEn2, -- active low
    AdrWr  => AdrWr,
    AdrRd1 => AdrRd1,
    AdrRd2 => AdrRd2,
    BankID => BankID,
    DataIn => Result,

    DataOut1 => DataAreg,
    DataOut2 => DataBreg
  );

-- PC ?
-- NPC ?
  DataA <= MemRead when (BypassRegA = '1')
      else Result when (BypassRegA2 = '1')
      else NPC when (AdrRd1 = REG_NPC)
      else DataAreg;

-- ImmXXX mux :
  DataB <= MemRead when (BypassRegB = '1')
      else NPC when (AdrRd2 = REG_NPC)
      else Result when (BypassRegB2 = '1')
      else Imm16 when (Form = FORM_RRI16)
      else Imm4 when ((Form = FORM_RI4) or (Form = FORM_EXT and AuxImm4='1'))
      else DataBreg;

--------------------------------------------------------------------------------
-- Multiply (creates bubbles before and after execution)

  assert (YASEP_MUL = 0) or (YASEP_MUL = 8)
    report "YASEP_MUL must be 0 or 8 for this version."
    severity failure;

  if (YASEP_MUL = 8) generate

    -- could be merged with a preliminary shift in IE or SHL ?
    MulMuxInput <= DataA( 7 downto 0) when (MulHiLo = '0')
              else DataA(15 downto 8);

    MUL8X4_LO:
      entity work.Mul8x4(A3P)
      port map(
        Clk   => clk,
        WrEn  => MulWrEn,
        MulEn => MulEn,
        DataA => DataB(3 downto 0),
        DataB => MulMuxInput,
        Val   => DataB(15 downto 8),
        DataOut => MulOutLo
      );

    MUL8X4_HI:
      entity work.Mul8x4(A3P)
      port map(
        Clk   => clk,
        WrEn  => MulWrEn,
        MulEn => MulEn,
        DataA => DataA(7 downto 4),
        DataB => MulMuxInput,
        Val   => DataB(15 downto 8),
        DataOut => MulOutHi
      );

-- HERE : add the pipeline bubble thing
-- MUX

-- signal delays :
  process(clk, add_ENr, IE_enr, SR_ENr, MulEnr, MulEn,
     Load, SignExt, OffsetA, IE_en, SR_EN, add_EN, SHL_en, ROP2_EN)
  begin
    if (clk'event and clk='1') then
      if ()

      Loadr    <= Load;
      SignExtr <= SignExt;
      OffsetAr <= OffsetA;
      IE_enr   <= IE_en;
      SR_ENr   <= SR_EN;
      add_ENr  <= add_EN or MulEnr;
      SHL_enr  <= SHL_en;
      ROP2_ENr <= ROP2_EN;
      DataAr   <= DataA;
      DataBr   <= DataB;
      MulEnr   <= MulEn;
    end if;
  end process;

  

  end generate;

  if (YASEP_MUL = 0) generate
    MulOutLo <= (others => '0');
    MulOutHi <= (others => '0');
  end generate;

-------------------------------------------------------------
-- ALU

ASU:
  entity work.ASUROP2_16(ActelProASIC3)
  generic map(
  	PIPELINED  => PIPELINED,
    YASEP_SIZE => YASEP_SIZE
  )
  port map(
    DataA  => DataA,
    DataB  => DataB,
    AuxLo  => MulOutLo,
    AuxHi  => MulOutHi,
    Addsub => Addsub,
    clk    => clk,
    add_EN => AddMulRop2,
    add_ENr => add_ENr,
    MulMux  => MulEnr,
    ROP2_S1 => ROP2_S1,
    ROP2_S2 => ROP2_S2,
    ROP2_X2 => ROP2_X2,

    ROP2_OUT => ROP2_OUT,
    ROP2_combine => ROP2_Combine, -- direct output, i'll fix it later.
    Sum    => Sum
  );

SHL:
  entity work.SHL_16(behaviour)
  generic map(
  	PIPELINED  => PIPELINED,
    YASEP_SIZE => YASEP_SIZE
  )
  port map(
    DataA      => DataA,
    SHAmount   => DataB(3 downto 0),
    clk        => clk,
    SHL_en     => SHL_en,
    Rotation   => Rotation,
    Direction  => Direction,
    Arithmetic => Arithmetic,

    ResultSHL  => ResSHL
  );

IE:
    entity work.IE_16(second)
    generic map(
      YASEP_SIZE => YASEP_SIZE
    )
    port map(
      DataA    => m5,
      DataB    => DataBr,
      Load     => Loadr,
      SignExt  => SignExtr,
      OffsetA  => OffsetAr,

      ResultIE => ResIE
    );


-- signal delays :
  process(clk, add_ENr, IE_enr, SR_ENr, MulEnr, MulEn,
     Load, SignExt, OffsetA, IE_en, SR_EN, add_EN, SHL_en, ROP2_EN)
  begin
    if (clk'event and clk='1') then
 -- 1st stage
      Loadr    <= Load;
      SignExtr <= SignExt;
      OffsetAr <= OffsetA;
      IE_enr   <= IE_en;
      SR_ENr   <= SR_EN;
      add_ENr  <= add_EN or MulEnr;
      SHL_enr  <= SHL_en;
      ROP2_ENr <= ROP2_EN;
      DataAr   <= DataA;
      DataBr   <= DataB;
      MulEnr   <= MulEn;
 -- 2nd stage
      add_ENrr <= add_ENr;
      IE_enrr  <= IE_enr;
      SR_ENrr  <= SR_ENr;
    end if;
  end process;


-- pipeline gates

  process(clk, SR_Read, SR_EN, ROP2_OUT, ResSHL, ROP2_ENr, SHL_enr, ResIE, IE_enr)
  begin
    if (PIPELINED=0) or (clk'event and clk='1' and SR_EN='1') then
      FF1 <= SR_Read;
    end if;

    if (PIPELINED=0) or (clk'event and clk='1' and (ROP2_ENr or SHL_enr)='1') then
      -- MUX4
      if (SHL_enr = '1') then
        FF2 <= ResSHL;
      else
        FF2 <= ROP2_OUT;
      end if;
    end if;

    if (PIPELINED=0) or (clk'event and clk='1' and IE_enr='1') then
      FF3 <= ResIE;
    end if;

  end process;


-- MUX :

  m5 <= MemRead when (MEM_enr = '1')
   else DataAr;

  m3 <= FF3 when (IE_enrr = '1')
   else MemRead;

  -- MUX2
  m2 <= FF1 when (SR_ENrr ='1')
   else FF2;

  -- MUX1
  m1 <= m3 when ((MEM_ENrr or IE_enrr)='1')
   else m2;

  -- MUX0
  Result <= Sum(YASEP_SIZE-1 downto 0) when (add_ENrr = '1')
   else m1;

  MemWrite <= Result; -- output the result to avoid "destructive optimisation"
  Carry_out <= Sum(YASEP_SIZE);

-- Write-back and decisions


end together;
