-- ALU_16.vhd
-- version 20090102
-- version 20090108 : adding MUL8x4
-- version 20090526 : added RegSet, this is now more a datapath than an ALU
-- This file gathers the separate units
-- and connects them so that latency is minimal.

--           _EN         _ENr                      _ENrr
-- ASU : MUL8x4---FF------------------------FFx---------------L8----\MUX0___Result
-- SR :        ---FF--[......extern......]--FF1-\_MUX2__m2__MUX1_m1_/(L9)
-- ROP2 :      ---FF------------L5-\_MUX4___FF2_/ (L7)      (L8)
-- SHL  :      ---FF------------L5-/ (L6)                   /
-- IE          ---FF---\MUX5-m5-IE---L2-----FF3-\_MUX3_m3__/
-- Mem          IOFF---/------------------------/

-- I should add : NPC, PC, 

library ieee;
use ieee.std_logic_1164.all;

entity ALU_16 is
  generic (
    YASEP_SIZE : integer := 16;
    PIPELINED  : integer := 1   -- pipelined when 1
  );
  port(
    Instruction : in std_logic_vector(31 downto 0); -- bit0 is the short/long flag

    MemRead, SR_Read : in std_logic_vector(YASEP_SIZE-1 downto 0);
    clk, SR_EN, MEM_ENr, MEM_ENrr, OffsetA, BypassRegA, BypassRegB,
    R7WrEn, R7RdEn1, R7RdEn2 : in std_logic;
    AdrWr, BankID : in std_logic_vector(3 downto 0);

    MemWrite : out std_logic_vector(YASEP_SIZE-1 downto 0);
    Carry_out, ROP2_Combine : out std_logic);
end ALU_16;

architecture together of ALU_16 is
  signal ROP2_OUT, ResSHL, ResIE, FF1, FF2, FF3, DataAr, DataBr,
     m1, m2, m3, m5, DataA, DataB, DataAreg, DataBreg, Result,
      Imm4, Imm16 : std_logic_vector(YASEP_SIZE-1 downto 0);
  signal Sum : std_logic_vector(YASEP_SIZE downto 0);
  signal MulOutLo, MulOutHi : std_logic_vector(11 downto 0);
  signal MulMuxInput : std_logic_vector(7 downto 0);
  signal AdrRd1, AdrRd2 : std_logic_vector(3 downto 0);
  signal GroupBits, FunctionBits : std_logic_vector(2 downto 0);
  signal Form : std_logic_vector(1 downto 0);
  signal IE_enr,  SR_ENr,  add_ENr,  MulEnr, SHL_enr, ROP2_ENr,
         IE_enrr, SR_ENrr, add_ENrr, Loadr, SignExtr, OffsetAr,
      add_EN, MulEn, SHL_en, IE_en, ROP2_EN, MulWrEn,
      AddMulRop2, MulHiLo, Rotation, Direction, Arithmetic,
      Load, SignExt, ROP2_S1, ROP2_S2, ROP2_X2, Addsub, AuxImm4 : std_logic;
begin

-- Instruction fields :
  Form <= Instruction(1 downto 0);
  GroupBits <= Instruction(4 downto 2);
  FunctionBits <= Instruction(7 downto 5);
  AdrRd1 <= Instruction(11 downto 8);
  AdrRd2 <= Instruction(15 downto 12);
  Imm4(AdrRd2'range) <= AdrRd2;
   Imm4(Imm4'left downto 4) <= (others => AdrRd2(3)); -- sign extension
  Imm16 <= Instruction(31 downto 16);

-- Extended form :
  AuxImm4 <= Instruction(16);

-- Groups :
  IE_en   <= '1' when (GroupBits = "000") else '0';
  SHL_en  <= '1' when (GroupBits = "001") else '0';
  ROP2_EN <= '1' when (GroupBits = "010") else '0';
  add_EN  <= '1' when (GroupBits = "011" or  MulEn = '1') else '0';
  MulEn   <= '1' when (GroupBits = "100") else '0';


-- multiply
  MulHiLo <= FunctionBits(0) and MulEn;
  MulWrEn <= '1' when (MulEn = '1' and FunctionBits="111") else '0';
-- SHL
  Direction  <= FunctionBits(0) and SHL_en;
  Rotation   <= FunctionBits(1) and SHL_en;
  Arithmetic <= FunctionBits(2) and SHL_en;
-- IE
  SignExt <= FunctionBits(0) and IE_en;
  Load    <= FunctionBits(1) and IE_en;


-- trick : reduce toggles and leave some unchanged values
  process(clk, GroupBits, FunctionBits)
    variable TmpDecode : std_logic_vector(3 downto 0);
  begin
    if (clk'event and clk='1') then
      if (GroupBits = "011") then
        Addsub <= FunctionBits(0) and (FunctionBits(1) or FunctionBits(2));
          -- dont substract on comparisons
      else
        case FunctionBits is                  --         ROP2_S1  ROP2_S2  ROP2_X2  Addsub
          when "000" => TmpDecode := "0000"; -- XOR        0        0        0       0
          when "001" => TmpDecode := "0001"; -- XORN       0        0        0       1
          when "010" => TmpDecode := "X100"; -- OR         X        1        0       0
          when "011" => TmpDecode := "X101"; -- ORN        X        1        0       1
          when "100" => TmpDecode := "X110"; -- NOR        X        1        1       0
          when "101" => TmpDecode := "1000"; -- AND        1        0        0       0
          when "110" => TmpDecode := "1001"; -- ANDN       1        0        0       1
          when "111" => TmpDecode := "1010"; -- NAND       1        0        1       0
        end case;
        Addsub  <= TmpDecode(0);
        ROP2_S1 <= TmpDecode(1);
        ROP2_S2 <= TmpDecode(2);
        ROP2_X2 <= TmpDecode(3);
      end if;
    end if;
  end process;

-------------------------------------------------------------------------------
-- The register set and the 2 main operands

R7:
  entity work.RegSet16(A3P)
  port map(
    Clk => clk,
    WrEn => R7WrEn, -- Low=Write
    RdEn1 => R7RdEN1,
    RdEn2 => R7RdEn2, -- active low
    AdrWr  => AdrWr,
    AdrRd1 => AdrRd1,
    AdrRd2 => AdrRd2,
    BankID => BankID,
    DataIn => Result,

    DataOut1 => DataAreg,
    DataOut2 => DataBreg
  );

  DataA <= MemRead when (BypassRegA = '1') else DataAreg;

-- ImmXXX mux :
  DataB <= MemRead when (BypassRegB = '1')
      else Imm16 when (Form = "10")
      else Imm4 when ((Form = "01") or (Form = "11" and AuxImm4='1'))
      else DataBreg;

--------------------------------------------------------------------------------
-- Multiply (creates a bubble)

-- could be merged with a preliminary shift in IE or SHL ?
MulMuxInput <= DataA( 7 downto 0) when (MulHiLo = '0')
          else DataA(15 downto 8);

MUL8X4_LO:
  entity work.Mul8x4(A3P)
  port map(
    Clk   => clk,
    WrEn  => MulWrEn,
    MulEn => MulEn,
    DataA => DataB(3 downto 0),
    DataB => MulMuxInput,
    Val   => DataB(15 downto 8),
    DataOut => MulOutLo
  );

MUL8X4_HI:
  entity work.Mul8x4(A3P)
  port map(
    Clk   => clk,
    WrEn  => MulWrEn,
    MulEn => MulEn,
    DataA => DataA(7 downto 4),
    DataB => MulMuxInput,
    Val   => DataB(15 downto 8),
    DataOut => MulOutHi
  );

  AddMulRop2 <= add_EN or ROP2_EN or MulEnr;

-------------------------------------------------------------
-- ALU


ASU:
  entity work.ASUROP2_16(ActelProASIC3)
  generic map(
  	PIPELINED  => PIPELINED,
    YASEP_SIZE => YASEP_SIZE
  )
  port map(
    DataA  => DataA,
    DataB  => DataB,
    AuxLo  => MulOutLo,
    AuxHi  => MulOutHi,
    Addsub => Addsub,
    clk    => clk,
    add_EN => AddMulRop2,
    add_ENr => add_ENr,
    MulMux  => MulEnr,
    ROP2_S1 => ROP2_S1,
    ROP2_S2 => ROP2_S2,
    ROP2_X2 => ROP2_X2,

    ROP2_OUT => ROP2_OUT,
    ROP2_combine => ROP2_Combine, -- direct output, i'll fix it later.
    Sum    => Sum
  );

SHL:
  entity work.SHL_16(behaviour)
  generic map(
  	PIPELINED  => PIPELINED,
    YASEP_SIZE => YASEP_SIZE
  )
  port map(
    DataA      => DataA,
    SHAmount   => DataB(3 downto 0),
    clk        => clk,
    SHL_en     => SHL_en,
    Rotation   => Rotation,
    Direction  => Direction,
    Arithmetic => Arithmetic,

    ResultSHL  => ResSHL
  );

IE:
    entity work.IE_16(second)
    generic map(
      YASEP_SIZE => YASEP_SIZE
    )
    port map(
      DataA    => m5,
      DataB    => DataBr,
      Load     => Loadr,
      SignExt  => SignExtr,
      OffsetA  => OffsetAr,

      ResultIE => ResIE
    );


-- signal delays :
  process(clk, add_ENr, IE_enr, SR_ENr, MulEnr, MulEn,
     Load, SignExt, OffsetA, IE_en, SR_EN, add_EN, SHL_en, ROP2_EN)
  begin
    if (clk'event and clk='1') then
 -- 1st stage
      Loadr    <= Load;
      SignExtr <= SignExt;
      OffsetAr <= OffsetA;
      IE_enr   <= IE_en;
      SR_ENr   <= SR_EN;
      add_ENr  <= add_EN or MulEnr;
      SHL_enr  <= SHL_en;
      ROP2_ENr <= ROP2_EN;
      DataAr   <= DataA;
      DataBr   <= DataB;
      MulEnr   <= MulEn;
 -- 2nd stage
      add_ENrr <= add_ENr;
      IE_enrr  <= IE_enr;
      SR_ENrr  <= SR_ENr;
    end if;
  end process;


-- pipeline gates

  process(clk, SR_Read, SR_EN, ROP2_OUT, ResSHL, ROP2_ENr, SHL_enr, ResIE, IE_enr)
  begin
    if (PIPELINED=0) or (clk'event and clk='1' and SR_EN='1') then
      FF1 <= SR_Read;
    end if;

    if (PIPELINED=0) or (clk'event and clk='1' and (ROP2_ENr or SHL_enr)='1') then
      -- MUX4
      if (SHL_enr = '1') then
        FF2 <= ResSHL;
      else
        FF2 <= ROP2_OUT;
      end if;
    end if;

    if (PIPELINED=0) or (clk'event and clk='1' and IE_enr='1') then
      FF3 <= ResIE;
    end if;

  end process;


-- MUX :

  m5 <= MemRead when (MEM_enr = '1')
   else DataAr;

  m3 <= FF3 when (IE_enrr = '1')
   else MemRead;

  -- MUX2
  m2 <= FF1 when (SR_ENrr ='1')
   else FF2;

  -- MUX1
  m1 <= m3 when ((MEM_ENrr or IE_enrr)='1')
   else m2;

  -- MUX0
  Result <= Sum(YASEP_SIZE-1 downto 0) when (add_ENrr = '1')
   else m1;

  MemWrite <= Result; -- output the result to avoid "destructive optimisation"
  Carry_out <= Sum(YASEP_SIZE);

-- Write-back and decisions



end together;
